import fs from 'node:fs/promises';
import cuffeo from 'cuffeo';
import oger from './inc/oger.mjs';
import lib from './inc/lib.mjs';
import config from '../config.json' with { type: 'json' };
import initials from '../data/initials.json' with { type: 'json' };

const conversations = new Map();
const bot = await new cuffeo(config.clients);
let last = false;
let playerlist = false;

const parseOutput = text => {
  text = text.trim().toLowerCase();
  return text.includes(', ')
    ? text.split(', ')
    : text.replace(/\d\. /g, '').split('\n');
};

bot.on('message', async e => {
  if(typeof e.message === 'undefined' || e.message.length <= 2)
    return;
  
  console.log(e.user.nick + ": " + e.message);
  const constr = `${e.network}::${e.channelid}`;

  if(e.message.match(/^!tags \d{1,}$/i)) {
    try {
      const id = e.message.substring(6);
      const f0ck = `https://f0ck.me/t/${id}.webp`;
      const blob = await (await fetch(f0ck)).arrayBuffer();

      const res = await fetch(`https://api.cloudflare.com/client/v4/accounts/${config.cf.id}/ai/run/${config.cf.model}`, {
        method: "POST",
        headers: {
          "Content-Type": "application/json",
          "Authorization": `Bearer ${config.cf.key}`
        },
        body: JSON.stringify({
          image: [ ...new Uint8Array(blob) ],
          prompt: "Generate five comma separated tags without colors for this image.",
          max_tokens: 128
        })
      });
      const json = await res.json();
      const tags = parseOutput(json.result.description);

      return await e.reply(`${f0ck}\n${JSON.stringify(tags)}`);
    } catch(err) {
      console.error(err);
      return;
    }
  }

  if(e.message.match(/^!messages/i)) {
    try {
      const cons = (await fs.readdir("./data/conversations")).filter(f => f.endsWith(".json"));
      let messages = 0;
      for(const con of cons) {
        messages += JSON.parse(await fs.readFile(`./data/conversations/${con}`, 'utf8')).length;
      }

      let output = [];
      output.push(`Nachrichten insgesamt: ${messages}`);

      if(cons.includes(constr+'.json')) {
        output.push(`Davon in diesem Kanal: ${JSON.parse(await fs.readFile(`./data/conversations/${constr}.json`, 'utf8')).length}`);
      }

      return await e.reply(output.join(' | '));
    } catch(err) {
      console.error(err);
      return;
    }
  }

  if(e.message.match(/^!gpustat/i)) {
    const gpustat = JSON.parse((await lib.exec('gpustat -a --json')).stdout);
    const gpu = gpustat.gpus[0];

    const output = [
      `${gpu['temperature.gpu']}°C, ${gpu['fan.speed']} rpm`,
      `${gpu['memory.used']} / ${gpu['memory.total']} MB`,
      `${gpu['power.draw']} / ${gpu['enforced.power.limit']} W`
    ].join(' | ');
    return await e.reply(`${gpustat.hostname}: ${gpu.name}\n${output}`);
  }

  if(e.message.match(/^!channelid/i)) {
    return await e.reply(e.channelid);
  }

  if(e.message.match(/^!reset/i)) {
    conversations.set(constr, []);
    await fs.writeFile(`./data/conversations/${constr}.json`, JSON.stringify([]));
    return await e.reply("Konversation wurde erfolgreich zurückgesetzt.");
  }

  let conv = conversations.has(constr) ? conversations.get(constr) : [];

  if(!conv.length) {
    try { // get conv from json
      const tmpcon = JSON.parse(await fs.readFile(`./data/conversations/${constr}.json`, 'utf8'));
      if(tmpcon) {
        conv = tmpcon;
        console.log(`Konversation von ${constr} geladen. ${tmpcon.length} Einträge.`);
      }
    } catch(err) {
      console.error(err);
      conversations.set(constr, conv);
    }
  }

  let tldr = false;
  if(e.message.match(/^!tldr/i)) {
    conv.push({
      role: 'user',
      content: `${e.user.nick} schreibt: schreibe eine kurze Zusammenfassung der letzten 50 Nachrichten.`
    });
    tldr = true;
  }

  if(!tldr) {
    conv.push({
      role: 'user',
      content: `${e.user.nick} schreibt: ${e.message.trim()}`
    });
    await fs.writeFile(`./data/conversations/${constr}.json`, JSON.stringify(conv));
    conversations.set(constr, conv);
  }

  if(!e.message?.match(/\bra+i+ne+r\b/i) && +e.self.me.id !== +e.raw?.reply_to_message?.from?.id && !tldr) {
    if(!config.initiative.includes(constr))
      return;

    let probability = 100;

    if(e.message.match(/.*\?$/))
      probability = 20;

    if(lib.rand(probability) !== 1)
      return;
  }


  // Rainer matched
  if(typeof e.sendChatAction !== 'undefined')
    await e.sendChatAction('typing');
  //while(last) await lib.sleep(500);

  const actcon = [{
    "role": "system",
    "content": initials[ initials[constr] ? constr : 'default' ].lore
  }, {
    "role": "system",
    "content": "Chatnachrichten von anderen sind immer folgendermaßen aufgebaut: \"%USERNAME% schreibt: %COMMAND%\". Du machst das aber nie."
  }];

  if(!tldr) {
    actcon.push(...conv.slice(-15));
  }
  else {
    actcon.push(...conv);
  }

  const opts = {
    method: 'POST',
    port: 5001,
    headers: {
      "Content-Type": "application/json",
      //"Authorization": `Bearer ${config.api.bearer}`
    },
    body: JSON.stringify({
      model: "em_german_leo_mistral.Q5_K_M.gguf",
      stream: false,
      max_tokens: 2048,
      top_k: 40,
      top_p: 0.92,
      temperature: 1.0,
      sampler_order: [6,0,1,3,4,2,5],
      rep_pen: 1.1,
      template: {
        chat: "chat-block",
        chat_message: "chat"
      },
      completion: "completion",
      context_size: 4096,
      use_fast_tokenizer: true,
      roles: {
        system: "<|im_start|> system",
        assistant: "<|im_start|> assistant",
        user: "<|im_start|> user"
      },
      stopwords: [
        "<|im_end|>",
        "### Response:"
      ],
      mmap: true,
      f16: true,
      messages: actcon
    })
  };

  last = e.user.nick;

  let res;

  try {
    res = await (await fetch(`http://${config.api.host}:${config.api.port}/v1/chat/completions`, opts)).json();
  } catch(err) {
    last = false;
    return await e.reply('Backend ist gerade offline. Sorreh.');
  }

  if(res?.choices?.[0]?.message) {
    if(res.choices[0].message.length <= 2)
      return await e.reply('ich habe dazu nichts zu sagen, du Birne.');

    res.choices[0].message.content = res.choices[0].message.content.split(/(###|---|ASSISTANT)/)[0];
    res.choices[0].message.content = res.choices[0].message.content.replace(/^.*( schreibt| antwortet)?\: /, '');

    conv.push(res.choices[0].message);
    conversations.set(constr, conv);
    await fs.writeFile(`./data/conversations/${constr}.json`, JSON.stringify(conv));

    console.log("Rainer: " + res.choices[0].message.content.trim());

    try {
      let orig = res.choices[0].message.content.trim();
      for(let i = 0, cl = orig.length; i < cl; i += 4000) {
        const msg = orig.substring(i, i + 4000);
        if(initials[constr] && initials[constr].langmod === 'oger') {
          await e.reply(oger(msg));
        }
        else {
          await e.reply(msg);
        }
      }
    } catch(err) {
      console.error(err);
      await e.reply('ups, kann ich nicht schreiben, wäre zu viel gewesen.');
    }
  }
  else {
    await e.reply('Hilfe, Fehler!');
  }

  last = false;
  return;
});

bot.on('error', err => {
  console.error(err);
});
